@@ -8,12 +8,18 @@ import (
 	"os/exec"
 	"path"
 	"path/filepath"
+	"strings"
 	"time"
 
+	"github.com/openshift/local-storage-operator/pkg/internal"
+
 	"github.com/ghodss/yaml"
+	"github.com/openshift/local-storage-operator/pkg/apis"
 	localv1 "github.com/openshift/local-storage-operator/pkg/apis/local/v1"
+	"github.com/prometheus/common/log"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/util/sets"
+	"k8s.io/client-go/kubernetes/scheme"
 	"k8s.io/klog"
 )
 
@@ -29,9 +35,9 @@ var (
 type DiskMaker struct {
 	configLocation  string
 	symlinkLocation string
-	apiClient       apiUpdater
+	apiClient       ApiUpdater
 	localVolume     *localv1.LocalVolume
-	eventSync       *eventReporter
+	eventSync       *EventReporter
 }
 
 type DiskLocation struct {
@@ -40,14 +46,22 @@ type DiskLocation struct {
 	diskID       string
 }
 
-// DiskMaker returns a new instance of DiskMaker
-func NewDiskMaker(configLocation, symLinkLocation string) *DiskMaker {
+// NewDiskMaker returns a new instance of DiskMaker
+func NewDiskMaker(configLocation, symLinkLocation string) (*DiskMaker, error) {
+	scheme := scheme.Scheme
+	apis.AddToScheme(scheme)
+	apiUpdater, err := NewAPIUpdater(scheme)
+	if err != nil {
+		log.Error(err, "failed to create new APIUpdater")
+		return &DiskMaker{}, err
+	}
+
 	t := &DiskMaker{}
 	t.configLocation = configLocation
 	t.symlinkLocation = symLinkLocation
-	t.apiClient = newAPIUpdater()
-	t.eventSync = newEventReporter(t.apiClient)
-	return t
+	t.apiClient = apiUpdater
+	t.eventSync = NewEventReporter(t.apiClient)
+	return t, nil
 }
 
 func (d *DiskMaker) loadConfig() (*DiskConfig, error) {
@@ -121,8 +135,8 @@ func (d *DiskMaker) symLinkDisks(diskConfig *DiskConfig) {
 	err = cmd.Run()
 	if err != nil {
 		msg := fmt.Sprintf("error running lsblk: %v", err)
-		e := newEvent(ErrorRunningBlockList, msg, "")
-		d.eventSync.report(e, d.localVolume)
+		e := NewEvent(ErrorRunningBlockList, msg, "")
+		d.eventSync.Report(e, d.localVolume)
 		klog.Errorf(msg)
 		return
 	}
@@ -137,46 +151,91 @@ func (d *DiskMaker) symLinkDisks(diskConfig *DiskConfig) {
 	allDiskIds, err := filepath.Glob(diskByIDPath)
 	if err != nil {
 		msg := fmt.Sprintf("error listing disks in /dev/disk/by-id: %v", err)
-		e := newEvent(ErrorListingDeviceID, msg, "")
-		d.eventSync.report(e, d.localVolume)
+		e := NewEvent(ErrorListingDeviceID, msg, "")
+		d.eventSync.Report(e, d.localVolume)
 		klog.Errorf(msg)
 		return
 	}
 
 	deviceMap, err := d.findMatchingDisks(diskConfig, deviceSet, allDiskIds)
 	if err != nil {
-		msg := fmt.Sprintf("eror finding matching disks: %v", err)
-		e := newEvent(ErrorFindingMatchingDisk, msg, "")
-		d.eventSync.report(e, d.localVolume)
+		msg := fmt.Sprintf("error finding matching disks: %v", err)
+		e := NewEvent(ErrorFindingMatchingDisk, msg, "")
+		d.eventSync.Report(e, d.localVolume)
 		klog.Errorf(msg)
 		return
 	}
 
 	if len(deviceMap) == 0 {
-		msg := "found empty matching device list"
-		e := newEvent(ErrorFindingMatchingDisk, msg, "")
-		d.eventSync.report(e, d.localVolume)
+		msg := ""
+		if len(diskConfig.Disks) == 0 {
+			// Note that this scenario shouldn't be possible, as diskConfig.Disks is a required attribute
+			msg = "devicePaths was left empty; this attribute must be defined in the LocalVolume spec"
+		} else {
+			// We go through and build up a set to display the differences between expected devices
+			// and invalid formatted entries (such as /tmp)
+			diff := sets.NewString()
+			for _, v := range diskConfig.Disks {
+				set := sets.NewString()
+				for _, devicePath := range v.DevicePaths {
+					set.Insert(devicePath)
+				}
+				diff = set.Difference(v.DeviceNames())
+			}
+			msg = strings.Join(diff.List(), ", ") + " was defined in devicePaths, but expected a path in /dev/"
+		}
+		e := NewEvent(ErrorFindingMatchingDisk, msg, "")
+		d.eventSync.Report(e, d.localVolume)
 		klog.Errorf(msg)
 		return
 	}
 
 	for storageClass, deviceArray := range deviceMap {
 		for _, deviceNameLocation := range deviceArray {
+			//
 			symLinkDirPath := path.Join(d.symlinkLocation, storageClass)
-			err := os.MkdirAll(symLinkDirPath, 0755)
+			baseDeviceName := filepath.Base(deviceNameLocation.diskNamePath)
+			symLinkPath := path.Join(symLinkDirPath, baseDeviceName)
+
+			// get PV creation lock which checks for existing symlinks to this device
+			pvLock, pvLocked, existingSymlinks, err := internal.GetPVCreationLock(
+				deviceNameLocation.diskNamePath,
+				d.symlinkLocation,
+			)
+			unlockFunc := func() {
+				err := pvLock.Unlock()
+				if err != nil {
+					klog.Errorf("failed to unlock device: %+v", err)
+				}
+			}
+			defer unlockFunc()
+			if len(existingSymlinks) > 0 { // already claimed, fail silently
+				e := NewEvent(DeviceSymlinkExists, "this device is already matched by another LocalVolume or LocalVolumeSet", symLinkPath)
+				d.eventSync.Report(e, d.localVolume)
+				unlockFunc()
+				continue
+			} else if err != nil || !pvLocked { // locking failed for some other reasion
+				klog.Errorf("not symlinking, could not get lock: %v", err)
+				unlockFunc()
+				continue
+			}
+
+			err = os.MkdirAll(symLinkDirPath, 0755)
 			if err != nil {
 				msg := fmt.Sprintf("error creating symlink dir %s: %v", symLinkDirPath, err)
-				e := newEvent(ErrorFindingMatchingDisk, msg, "")
-				d.eventSync.report(e, d.localVolume)
+				e := NewEvent(ErrorFindingMatchingDisk, msg, symLinkPath)
+				d.eventSync.Report(e, d.localVolume)
+				unlockFunc()
 				klog.Errorf(msg)
 				continue
 			}
-			baseDeviceName := filepath.Base(deviceNameLocation.diskNamePath)
-			symLinkPath := path.Join(symLinkDirPath, baseDeviceName)
+
 			if fileExists(symLinkPath) {
+				unlockFunc()
 				klog.V(4).Infof("symlink %s already exists", symLinkPath)
 				continue
 			}
+
 			var symLinkErr error
 			if deviceNameLocation.diskID != "" {
 				klog.V(3).Infof("symlinking to %s to %s", deviceNameLocation.diskID, symLinkPath)
@@ -185,17 +244,17 @@ func (d *DiskMaker) symLinkDisks(diskConfig *DiskConfig) {
 				klog.V(3).Infof("symlinking to %s to %s", deviceNameLocation.diskNamePath, symLinkPath)
 				symLinkErr = os.Symlink(deviceNameLocation.diskNamePath, symLinkPath)
 			}
-
 			if symLinkErr != nil {
-				msg := fmt.Sprintf("error creating symlink %s: %v", symLinkPath, err)
-				e := newEvent(ErrorFindingMatchingDisk, msg, deviceNameLocation.diskNamePath)
-				d.eventSync.report(e, d.localVolume)
+				msg := fmt.Sprintf("error creating symlink %s: %v", symLinkPath, symLinkErr)
+				e := NewEvent(ErrorFindingMatchingDisk, msg, deviceNameLocation.diskNamePath)
+				d.eventSync.Report(e, d.localVolume)
 				klog.Errorf(msg)
 			}
-
+			// unlock before proceeding
+			unlockFunc()
 			successMsg := fmt.Sprintf("found matching disk %s", baseDeviceName)
-			e := newSuccessEvent(FoundMatchingDisk, successMsg, deviceNameLocation.diskNamePath)
-			d.eventSync.report(e, d.localVolume)
+			e := NewSuccessEvent(FoundMatchingDisk, successMsg, deviceNameLocation.diskNamePath)
+			d.eventSync.Report(e, d.localVolume)
 		}
 	}
 
@@ -229,9 +288,27 @@ func (d *DiskMaker) findMatchingDisks(diskConfig *DiskConfig, deviceSet sets.Str
 				addDiskToMap(storageClass, matchedDeviceID, diskName)
 				continue
 			} else {
-				msg := fmt.Sprintf("unable to find matching disk %v", diskName)
-				e := newEvent(ErrorFindingMatchingDisk, msg, diskName)
-				d.eventSync.report(e, d.localVolume)
+				if !fileExists(diskName) {
+					msg := fmt.Sprintf("no file exists for specified device %v", diskName)
+					klog.Errorf(msg)
+					continue
+				}
+				fileMode, err := os.Stat(diskName)
+				if err != nil {
+					klog.Errorf("error attempting to examine %v, %v", diskName, err)
+					continue
+				}
+				msg := ""
+				switch mode := fileMode.Mode(); {
+				case mode.IsDir():
+					msg = fmt.Sprintf("unable to use directory %v for local storage. Use an existing block device.", diskName)
+				case mode.IsRegular():
+					msg = fmt.Sprintf("unable to use regular file %v for local storage. Use an existing block device.", diskName)
+				default:
+					msg = fmt.Sprintf("unable to find matching disk %v", diskName)
+				}
+				e := NewEvent(ErrorFindingMatchingDisk, msg, diskName)
+				d.eventSync.Report(e, d.localVolume)
 				klog.Errorf(msg)
 			}
 		}
@@ -242,8 +319,8 @@ func (d *DiskMaker) findMatchingDisks(diskConfig *DiskConfig, deviceSet sets.Str
 			matchedDeviceID, matchedDiskName, err := d.findDeviceByID(deviceID)
 			if err != nil {
 				msg := fmt.Sprintf("unable to add disk-id %s to local disk pool: %v", deviceID, err)
-				e := newEvent(ErrorFindingMatchingDisk, msg, deviceID)
-				d.eventSync.report(e, d.localVolume)
+				e := NewEvent(ErrorFindingMatchingDisk, msg, deviceID)
+				d.eventSync.Report(e, d.localVolume)
 				klog.Errorf(msg)
 				continue
 			}
@@ -261,7 +338,7 @@ func (d *DiskMaker) findMatchingDisks(diskConfig *DiskConfig, deviceSet sets.Str
 func (d *DiskMaker) findDeviceByID(deviceID string) (string, string, error) {
 	diskDevPath, err := filepath.EvalSymlinks(deviceID)
 	if err != nil {
-		return "", "", fmt.Errorf("unable to find device with id %s", deviceID)
+		return "", "", fmt.Errorf("unable to find device at path %s: %v", deviceID, err)
 	}
 	return deviceID, diskDevPath, nil
 }
